# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations

from abc import ABC, abstractmethod
from queue import Queue
from typing import TYPE_CHECKING, List, Optional, Union

from sila2.server import FeatureImplementationBase, MetadataDict

from .ufcontroller_types import (
    CloseArm_Responses,
    CloseGripper_Responses,
    InitializeArm_Responses,
    MoveJoints_Responses,
    MoveToStoredPosition_Responses,
    OpenGripper_Responses,
    ResetArm_Responses,
    SetManualMode_Responses,
    StorePosition_Responses,
)

if TYPE_CHECKING:

    from ...server import Server


class UFControllerBase(FeatureImplementationBase, ABC):
    parent_server: Server

    _JointAngles_producer_queue: Queue[Union[List[float], Exception]]
    _JointAngles_current_value: List[float]

    _GripperPosition_producer_queue: Queue[Union[float, Exception]]
    _GripperPosition_current_value: float

    _JointVelocities_producer_queue: Queue[Union[List[float], Exception]]
    _JointVelocities_current_value: List[float]

    def __init__(self, parent_server: Server):
        """
        This controller is intended to be a SiLA2 compliant controller for Ufactory robot arms.
        """
        super().__init__(parent_server=parent_server)

        self._JointAngles_producer_queue = Queue()

        self._GripperPosition_producer_queue = Queue()

        self._JointVelocities_producer_queue = Queue()

    @abstractmethod
    def get_ArmIPAddress(self, *, metadata: MetadataDict) -> str:
        """
        The IP Address of the connected robotic arm.

        :param metadata: The SiLA Client Metadata attached to the call
        :return: The IP Address of the connected robotic arm.
        """

    def update_JointAngles(self, JointAngles: List[float], queue: Optional[Queue[List[float]]] = None) -> None:
        """
        Return the joint positions to the client.

        This method updates the observable property 'JointAngles'.

        :param queue: The queue to send updates to. If None, the default Queue will be used.
        """
        if queue is None:
            queue = self._JointAngles_producer_queue
            self._JointAngles_current_value = JointAngles
        queue.put(JointAngles)

    def JointAngles_on_subscription(self, *, metadata: MetadataDict) -> Optional[Queue[List[float]]]:
        """
        Return the joint positions to the client.

        This method is called when a client subscribes to the observable property 'JointAngles'

        :param metadata: The SiLA Client Metadata attached to the call
        :return: Optional `Queue` that should be used for updating this property.
            If None, the default Queue will be used.
        """

    def abort_JointAngles_subscriptions(self, error: Exception, queue: Optional[Queue[List[float]]] = None) -> None:
        """
        Return the joint positions to the client.

        This method aborts subscriptions to the observable property 'JointAngles'.

        :param error: The Exception to be sent to the subscribing client.
            If it is no DefinedExecutionError or UndefinedExecutionError, it will be wrapped in an UndefinedExecutionError.
        :param queue: The queue to abort. If None, the default Queue will be used.
        """
        if queue is None:
            queue = self._JointAngles_producer_queue
        queue.put(error)

    @property
    def current_JointAngles(self) -> List[float]:
        try:
            return self._JointAngles_current_value
        except AttributeError:
            raise AttributeError("Observable property JointAngles has never been set")

    def update_GripperPosition(self, GripperPosition: float, queue: Optional[Queue[float]] = None) -> None:
        """
        Returns the current gripper position.

        This method updates the observable property 'GripperPosition'.

        :param queue: The queue to send updates to. If None, the default Queue will be used.
        """
        if queue is None:
            queue = self._GripperPosition_producer_queue
            self._GripperPosition_current_value = GripperPosition
        queue.put(GripperPosition)

    def GripperPosition_on_subscription(self, *, metadata: MetadataDict) -> Optional[Queue[float]]:
        """
        Returns the current gripper position.

        This method is called when a client subscribes to the observable property 'GripperPosition'

        :param metadata: The SiLA Client Metadata attached to the call
        :return: Optional `Queue` that should be used for updating this property.
            If None, the default Queue will be used.
        """

    def abort_GripperPosition_subscriptions(self, error: Exception, queue: Optional[Queue[float]] = None) -> None:
        """
        Returns the current gripper position.

        This method aborts subscriptions to the observable property 'GripperPosition'.

        :param error: The Exception to be sent to the subscribing client.
            If it is no DefinedExecutionError or UndefinedExecutionError, it will be wrapped in an UndefinedExecutionError.
        :param queue: The queue to abort. If None, the default Queue will be used.
        """
        if queue is None:
            queue = self._GripperPosition_producer_queue
        queue.put(error)

    @property
    def current_GripperPosition(self) -> float:
        try:
            return self._GripperPosition_current_value
        except AttributeError:
            raise AttributeError("Observable property GripperPosition has never been set")

    def update_JointVelocities(self, JointVelocities: List[float], queue: Optional[Queue[List[float]]] = None) -> None:
        """
        Returns the velocities of each joint in real-time.

        This method updates the observable property 'JointVelocities'.

        :param queue: The queue to send updates to. If None, the default Queue will be used.
        """
        if queue is None:
            queue = self._JointVelocities_producer_queue
            self._JointVelocities_current_value = JointVelocities
        queue.put(JointVelocities)

    def JointVelocities_on_subscription(self, *, metadata: MetadataDict) -> Optional[Queue[List[float]]]:
        """
        Returns the velocities of each joint in real-time.

        This method is called when a client subscribes to the observable property 'JointVelocities'

        :param metadata: The SiLA Client Metadata attached to the call
        :return: Optional `Queue` that should be used for updating this property.
            If None, the default Queue will be used.
        """

    def abort_JointVelocities_subscriptions(self, error: Exception, queue: Optional[Queue[List[float]]] = None) -> None:
        """
        Returns the velocities of each joint in real-time.

        This method aborts subscriptions to the observable property 'JointVelocities'.

        :param error: The Exception to be sent to the subscribing client.
            If it is no DefinedExecutionError or UndefinedExecutionError, it will be wrapped in an UndefinedExecutionError.
        :param queue: The queue to abort. If None, the default Queue will be used.
        """
        if queue is None:
            queue = self._JointVelocities_producer_queue
        queue.put(error)

    @property
    def current_JointVelocities(self) -> List[float]:
        try:
            return self._JointVelocities_current_value
        except AttributeError:
            raise AttributeError("Observable property JointVelocities has never been set")

    @abstractmethod
    def OpenGripper(self, *, metadata: MetadataDict) -> OpenGripper_Responses:
        """
        Open the gripper as far as possible.


        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def CloseGripper(self, *, metadata: MetadataDict) -> CloseGripper_Responses:
        """
        Close the gripper until it encounters a counter force.


        :param metadata: The SiLA Client Metadata attached to the call

        """

    @abstractmethod
    def InitializeArm(self, *, metadata: MetadataDict) -> InitializeArm_Responses:
        """
        Initialize the robotic arm and set it in a ready state.


        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - Success: Indicates whether the initialization was successful.


        """

    @abstractmethod
    def CloseArm(self, *, metadata: MetadataDict) -> CloseArm_Responses:
        """
        Safely shut down and disconnect the robotic arm.


        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - Success: Indicates whether the shutdown was successful.


        """

    @abstractmethod
    def ResetArm(self, *, metadata: MetadataDict) -> ResetArm_Responses:
        """
        Resets the arm by closing and reopening the connection.


        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - Success: Indicates whether the reset was successful.


        """

    @abstractmethod
    def SetManualMode(self, *, metadata: MetadataDict) -> SetManualMode_Responses:
        """
        Switches the robotic arm to manual control mode.


        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - Success: Indicates whether the arm was successfully switched to manual mode.


        """

    @abstractmethod
    def MoveJoints(self, JointAngles: List[float], *, metadata: MetadataDict) -> MoveJoints_Responses:
        """
        Move the robot arm to a new position


        :param JointAngles: The list of joint angles with their names and new positions.

        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - MoveJointsResponses: Response indicating the success of the move and the current joint angles.


        """

    @abstractmethod
    def StorePosition(self, *, metadata: MetadataDict) -> StorePosition_Responses:
        """
        Add the current position to the list of stored positions. Will return the number.


        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - PositionIdentifier: The position in the list of stored poses


        """

    @abstractmethod
    def MoveToStoredPosition(
        self, PositionIdentifier: int, *, metadata: MetadataDict
    ) -> MoveToStoredPosition_Responses:
        """
        Move the robot arm to a previously stored position


        :param PositionIdentifier: The position in the list of stored poses as previously returned by store position

        :param metadata: The SiLA Client Metadata attached to the call

        """



# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations
import asyncio
from .gripper_base import ArduinoGripper
from xarm.wrapper import XArmAPI
from typing import TYPE_CHECKING, Any, Dict, List
from sila2.server import MetadataDict

from sila2.server import MetadataDict

from ..generated.ufcontroller import (
    CloseArm_Responses,
    CloseGripper_Responses,
    InitializeArm_Responses,
    MoveJoints_Responses,
    MoveToStoredPosition_Responses,
    OpenGripper_Responses,
    ResetArm_Responses,
    SetManualMode_Responses,
    StorePosition_Responses,
    UFControllerBase,
)

"""
 Change the IP as neededy
"""
arm_ip = '192.168.1.160'

if TYPE_CHECKING:
    from ..server import Server


class UFControllerImpl(UFControllerBase):
    def __init__(self, parent_server: Server) -> None:
        gripper = ArduinoGripper()
        arm = XArmAPI(arm_ip)

        # Connect to the arm
        arm.connect()

        # Print some basic info about the arm
        print("Connected to xArm with IP: ", arm_ip)
        print("Arm State: ", arm.get_state())
        print("Current Position: ", arm.get_position())

        # Enable motion on the arm
        arm.motion_enable(enable=True)

        # Set the mode to position control
        arm.set_mode(0)

        # Set the state to be ready to receive commands
        arm.set_state(0)

        self.arm = arm
        self.gripper = gripper
        self.manual_mode_enabled = False  # Initialize manual mode state

        # Initialize position storage and periodic variables
        self.position_storage: Dict[int, List[float]] = {}
        self.periodic_variables: Dict[str, Any] = {}

        super().__init__(parent_server=parent_server)
    
    @property
    def JointAngles(self) -> List[float]:
        """Property representing the current joint angles of the arm."""
        return self.arm.get_servo_angle()

    async def monitor_joint_angles(self):
        """Background task to monitor joint angles and notify observers if changes occur."""
        while True:
            await asyncio.sleep(1)  # Poll every second
            current_angles = self.JointAngles
            if current_angles != self._current_joint_angles:
                self._current_joint_angles = current_angles
                self.notify_observers(current_angles)
                
    def notify_observers(self, joint_angles: List[float]):
        """Notify clients of changes to JointAngles."""
        # Example print statement to simulate notification
        print(f"JointAngles updated: {joint_angles}")
        # In a real implementation, you could send a message or trigger an event to notify clients

    def get_ArmIPAddress(self, *, metadata: MetadataDict) -> str:
        """Retrieve the IP address of the robotic arm."""
        return arm_ip

    def OpenGripper(self, *, metadata: MetadataDict) -> OpenGripper_Responses:
        """Open the gripper attached to the robotic arm."""
        self.gripper.send_command(command='o')
        return OpenGripper_Responses(success=True)

    def CloseGripper(self, *, metadata: MetadataDict) -> CloseGripper_Responses:
        """Close the gripper attached to the robotic arm."""
        self.gripper.send_command(command='c')
        return CloseGripper_Responses(success=True)

    def InitializeArm(self, *, metadata: MetadataDict) -> InitializeArm_Responses:
        """Initialize the robotic arm to its default state."""
        self.arm.motion_enable(enable=True)
        self.arm.set_mode(0)
        self.arm.set_state(0)
        self.arm.move_gohome()  # Move to home position
        print("Arm initialized and moved to home position.")
        return InitializeArm_Responses(success=True)

    def CloseArm(self, *, metadata: MetadataDict) -> CloseArm_Responses:
        """Disconnect and disable the robotic arm."""
        self.arm.disconnect()
        print("Arm connection closed.")
        return CloseArm_Responses(success=True)

    def ResetArm(self, *, metadata: MetadataDict) -> ResetArm_Responses:
        """Reset the robotic arm, usually after an error or emergency stop."""
        self.arm.reset(wait=True)
        print("Arm reset.")
        return ResetArm_Responses(success=True)

    def SetManualMode(self, *, metadata: MetadataDict) -> SetManualMode_Responses:
        """
        Toggle manual mode. If manual mode is off, turn it on.
        If manual mode is on, turn it off.
        """
        if not self.manual_mode_enabled:
            # Enable manual mode
            self.arm.set_mode(mode=2, detection_param=0)  # Set manual mode
            self.arm.set_state(0)  # Make the arm ready
            print("Manual mode enabled.")
            self.manual_mode_enabled = True
        else:
            # Disable manual mode
            self.arm.motion_enable(enable=True)
            self.arm.set_mode(0)  # Set to default mode
            self.arm.set_state(0)  # Make the arm ready
            print("Manual mode disabled.")
            self.manual_mode_enabled = False

        return SetManualMode_Responses(success=True)

    def MoveJoints(self, JointAngles: List[float], *, metadata: MetadataDict) -> MoveJoints_Responses:
        """Move the robotic arm to specified joint angles."""
        print("Moving to position: ", JointAngles)
        self.arm.set_servo_angle(angle=JointAngles, wait=True)
        return MoveJoints_Responses(JointAngles)

    def StorePosition(self, *, metadata: MetadataDict) -> StorePosition_Responses:
        """Store the current position of the robotic arm."""
        current_position = self.arm.get_servo_angle()
        position_id = len(self.position_storage) + 1
        self.position_storage[position_id] = current_position
        print(f"Position {position_id} stored: {current_position}")
        return StorePosition_Responses(PositionIdentifier=position_id)

    def MoveToStoredPosition(
        self, PositionIdentifier: int, *, metadata: MetadataDict
    ) -> MoveToStoredPosition_Responses:
        """Move the arm to a previously stored position using its identifier."""
        if PositionIdentifier not in self.position_storage:
            print(f"Position ID {PositionIdentifier} not found.")
            return MoveToStoredPosition_Responses(success=False)
        
        stored_position = self.position_storage[PositionIdentifier]
        self.arm.set_servo_angle(angle=stored_position, wait=True)
        print(f"Moved to stored position {PositionIdentifier}: {stored_position}")
        return MoveToStoredPosition_Responses(success=True)



# Generated by sila2.code_generator; sila2.__version__: 0.12.2
from __future__ import annotations

from typing import List, NamedTuple


class OpenGripper_Responses(NamedTuple):

    pass


class CloseGripper_Responses(NamedTuple):

    pass


class InitializeArm_Responses(NamedTuple):

    Success: bool
    """
    Indicates whether the initialization was successful.
    """


class CloseArm_Responses(NamedTuple):

    Success: bool
    """
    Indicates whether the shutdown was successful.
    """


class ResetArm_Responses(NamedTuple):

    Success: bool
    """
    Indicates whether the reset was successful.
    """


class SetManualMode_Responses(NamedTuple):

    Success: bool
    """
    Indicates whether the arm was successfully switched to manual mode.
    """


class MoveJoints_Responses(NamedTuple):

    MoveJointsResponses: List[float]
    """
    Response indicating the success of the move and the current joint angles.
    """


class StorePosition_Responses(NamedTuple):

    PositionIdentifier: int
    """
    The position in the list of stored poses
    """


class MoveToStoredPosition_Responses(NamedTuple):

    pass
